from __future__ import annotations

import abc
import inspect
from dataclasses import dataclass
from functools import cached_property
from typing import Optional, Dict, List, Sequence, Union, Callable, Any, Iterator

from lungo.lexer import Position


class CallStack:
    """Represents function call stack."""

    @dataclass(frozen=True)
    class Frame:
        pos: Position

    def __init__(self, pos: Optional[Position] = None):
        self._frames: List[CallStack.Frame] = []
        if pos is not None:
            self._frames.append(CallStack.Frame(pos))

    def push(self, frame: Frame):
        self._frames.append(frame)

    def pop(self) -> Frame:
        return self._frames.pop()

    def __iter__(self) -> Iterator[Frame]:
        """Iterate over frames."""
        for frame in self._frames:
            yield frame


class ScopeError(Exception):
    """Parent class for Scope errors."""


class UndefinedName(ScopeError):
    """Name is not defined in the given scope."""

    def __init__(self, name: str):
        super().__init__(f"Name is not defined in the scope: {name}")
        self.name: str = name


class NameRedefined(ScopeError):
    """Attempt to define name already existing in the scope."""

    def __init__(self, name: str):
        super().__init__(f"Cannot redefine variable name: {name}")
        self.name: str = name


class Scope:
    """Hierarchy of lexical scopes, a table of known variable names."""

    def __init__(self, symbols: Optional[SymbolTable] = None, parent: Optional[Scope] = None):
        symbols = symbols or {}
        self._symbols: SymbolTable = {**symbols}
        self.parent: Optional[Scope] = parent

    def assign(self, name: str, value):
        if name in self._symbols:
            self._symbols[name] = value
        elif self.parent is not None:
            self.parent.assign(name, value)
        else:
            raise UndefinedName(name)

    def define(self, name: str, value):
        if name in self._symbols:
            raise NameRedefined(name)
        self._symbols[name] = value

    def deref(self, name: str):
        if name in self._symbols:
            return self._symbols[name]
        elif self.parent is not None:
            return self.parent.deref(name)
        else:
            raise UndefinedName(name)

    def nested(self, symbols: Optional[SymbolTable] = None) -> Scope:
        return Scope(symbols=symbols, parent=self)


class ExecutionContext:
    def __init__(self, scope: Scope, stack: CallStack):
        self.scope: Scope = scope
        self.stack: CallStack = stack

    def nested(self, symbols: Optional[SymbolTable] = None) -> ExecutionContext:
        """Create execution context with nested lexical scope."""
        return ExecutionContext(self.scope.nested(symbols), self.stack)

    def push(self, pos: Position, scope: Scope):
        """Push another stack frame and prepare function scope."""
        self.stack.push(CallStack.Frame(pos))
        return ExecutionContext(scope, self.stack)


class ExecutionError(Exception):
    def __init__(self, message, stack: CallStack, pos: Position):
        super().__init__(message)
        self.stack = stack
        self.pos = pos


class Value:
    """Parent class for all values."""
    type: Type

    @cached_property
    def attrs(self) -> AttributeHolder:
        """Get attributes."""
        attrs = self._make_attributes()
        self._define_attributes(attrs)
        return attrs

    def _make_attributes(self) -> AttributeHolder:
        """Make attributes."""
        return AttributeHolder(self, can_set_unknown=False)

    def _define_attributes(self, attrs: AttributeHolder):
        """Define existing attributes."""
        attrs.define(type=Const(self.type))

    def plus(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Plus operator."""
        raise ExecutionError(f"Can't add to {self.type.name}", context.stack, pos)

    def minus(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Minus operator."""
        raise ExecutionError(f"Can't subtract from {self.type.name}", context.stack, pos)

    def mul(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Multiplication operator."""
        raise ExecutionError(f"Can't multiply {self.type.name}", context.stack, pos)

    def div(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Division operator."""
        raise ExecutionError(f"Can't divide {self.type.name}", context.stack, pos)

    def le(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Lesser-equal operator."""
        raise ExecutionError(f"Can't check {self.type.name} is lesser than or equal to other value", context.stack, pos)

    def lt(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Lesser-than operator."""
        raise ExecutionError(f"Can't check if {self.type.name} is lesser than other value", context.stack, pos)

    def ge(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Greater-equal operator."""
        raise ExecutionError(f"Can't check if {self.type.name} greater or equal to other value", context.stack, pos)

    def gt(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Greater-than operator."""
        raise ExecutionError(f"Can't check if {self.type.name} is greater than other value", context.stack, pos)

    def eq(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Equality operator."""
        return Bool.instance(self is other)

    def and_(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Logical-and operator."""
        if self.is_truthy():
            return other
        return self

    def or_(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Logical-or operator."""
        if self.is_truthy():
            return self
        return other

    def not_(self, context: ExecutionContext, pos: Position) -> Value:
        """Logical-not unary operator."""
        return Bool.instance(not self.is_truthy())

    def neg(self, context: ExecutionContext, pos: Position) -> Value:
        """Negation unary operator."""
        raise ExecutionError(f"Can't negate {self.type.name}", context.stack, pos)

    def getitem(self, key: Value, context: ExecutionContext, pos: Position) -> Value:
        """Get collection element by index."""
        raise ExecutionError(f"Can't get item from {self.type.name}", context.stack, pos)

    def setitem(self, key: Value, value: Value, context: ExecutionContext, pos: Position) -> Value:
        """Set collection item."""
        raise ExecutionError(f"Can't set item on the {self.type.name}", context.stack, pos)

    def call(self, arguments: Sequence[Value], context: ExecutionContext, pos: Position) -> Value:
        """Invoke-function operator."""
        raise ExecutionError(f"{self.type.name} is not callable", context.stack, pos)

    def getattr(self, name: str, context: ExecutionContext, pos: Position) -> Value:
        """Get-attribute operator."""
        return self.attrs.get_value(name, context, pos)

    def setattr(self, name: str, value: Value, context: ExecutionContext, pos: Position) -> Value:
        """Set-attribute operator."""
        return self.attrs.set_value(name, value, context, pos)

    def hasattr(self, name: str, context: ExecutionContext, pos: Position) -> Bool:
        """Check if attribute exists."""
        return Bool.instance(self.attrs.exists(name))

    def to_bool(self, context: ExecutionContext, pos: Position) -> Bool:
        """Converto to boolean."""
        return Bool.instance(self.is_truthy())

    def to_num(self, context: ExecutionContext, pos: Position) -> Number:
        """Convert to number."""
        raise ExecutionError(f"Can't convert {self.type.name} to {NumberType.name}", context.stack, pos)

    def to_str(self, context: ExecutionContext, pos: Position) -> String:
        """Convert to string."""
        raise ExecutionError(f"Can't convert {self.type.name} to {StringType.name}", context.stack, pos)

    @staticmethod
    def is_truthy() -> bool:
        """Check if the value is truthy."""
        return True


SymbolTable = Dict[str, Value]


class Attribute(abc.ABC):

    @abc.abstractmethod
    def get(self) -> Value:
        """Get attribute value."""

    @abc.abstractmethod
    def set(self, value: Value) -> Value:
        """Set attribute value."""

    @abc.abstractmethod
    def can_set(self) -> bool:
        """Check if setting the attribute is allowed."""


class SetAttributeError(Exception):
    """Can't set attribute on object."""


class GetAttributeError(Exception):
    """Can't get attribute of object."""


class Const(Attribute):
    def __init__(self, value: Value):
        self.value: Value = value

    def get(self) -> Value:
        return self.value

    def set(self, value: Value) -> Value:
        raise SetAttributeError()

    def can_set(self) -> bool:
        return False


Getter = Callable[[], Value]
Setter = Callable[[Value], Value]


class Property(Attribute):
    def __init__(self, getter: Getter, setter: Optional[Setter] = None):
        self._getter: Getter = getter
        self._setter: Optional[Setter] = setter

    def get(self) -> Value:
        return self._getter()

    def set(self, value: Value) -> Value:
        if self._setter is None:
            raise SetAttributeError()
        return self._setter(value)

    def can_set(self) -> bool:
        return self._setter is not None


AttributeValue = Union[Attribute, Value]


class AttributeHolder:
    """Attribute holder."""

    def __init__(self, instance: Value, can_set_unknown: bool = False):
        self._attributes: Dict[str, AttributeValue] = {}
        self._instance: Value = instance
        self._can_set_unknown: bool = can_set_unknown

    def get_value(self, name: str, context: ExecutionContext, pos: Position) -> Value:
        """Get attribute value."""
        if name not in self._attributes:
            raise ExecutionError(f"'{self._instance.type}' object has no attribute '{name}'", context.stack, pos)
        attribute_value = self._attributes[name]
        if isinstance(attribute_value, Value):
            return attribute_value
        if isinstance(attribute_value, Attribute):
            try:
                return attribute_value.get()
            except GetAttributeError:
                raise ExecutionError(
                    f"Can't get attribute '{name}' of '{self._instance.type}' object", context.stack, pos)

    def set_value(self, name: str, value: Value, context: ExecutionContext, pos: Position) -> Value:
        """Set attribute value."""
        if name not in self._attributes and not self._can_set_unknown:
            raise ExecutionError(f"Can't set attribute '{name}' on '{self._instance.type}'", context.stack, pos)
        if name not in self._attributes:
            self._attributes[name] = value
            return value
        attribute_value = self._attributes[name]
        if isinstance(attribute_value, Value):
            self._attributes[name] = value
            return value
        if not attribute_value.can_set():
            raise ExecutionError(f"Can't set attribute '{name}' on '{self._instance.type}'", context.stack, pos)
        try:
            attribute_value.set(value)
        except SetAttributeError:
            raise ExecutionError(f"Can't set attribute '{name}' on '{self._instance.type}'", context.stack, pos)

    def exists(self, name: str) -> bool:
        """Check if the attribute exists."""
        return name in self._attributes

    def define(self, **declarations: AttributeValue):
        """Define attributes."""
        self._attributes.update(declarations)


class Type(Value):
    """Value type."""
    name: str = "Type"

    def __init__(self):
        if type(self) is Type:
            self.type = self
        else:
            self.type = Type.instance()

    def _define_attributes(self, attrs: AttributeHolder):
        super()._define_attributes(attrs)
        attrs.define(name=Const(from_py(self.name)))

    @classmethod
    def instance(cls) -> "Type":
        if "__instance" not in cls.__dict__:
            setattr(cls, "__instance", cls())
        return getattr(cls, "__instance")

    def __repr__(self):
        return self.name


class BoolType(Type):
    """Type for boolean values."""
    name: str = "Bool"


class Bool(Value):
    """Boolean value."""
    type = BoolType.instance()

    true: Bool
    false: Bool

    def __init__(self, value: bool):
        self.value: bool = value

    @staticmethod
    def instance(value: bool) -> Bool:
        return Bool.true if value else Bool.false

    def eq(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Equality operator."""
        if isinstance(other, Bool):
            return Bool.instance(self.value == other.value)
        return Bool.false

    def to_bool(self, context: ExecutionContext, pos: Position) -> Bool:
        """Converto to boolean."""
        return self

    def to_num(self, context: ExecutionContext, pos: Position) -> Number:
        """Convert to number."""
        if self.value:
            return Number(1)
        return Number(0)

    def to_str(self, context: ExecutionContext, pos: Position) -> String:
        return String(str(self))

    def is_truthy(self) -> bool:
        return self.value

    def __str__(self):
        if self.value:
            return "true"
        return "false"

    def __repr__(self):
        return f"{self.type.name}({str(self)})"


Bool.true = Bool(True)
Bool.false = Bool(False)


class NilType(Type):
    """Type for 'nil' value."""
    name: str = 'Nil'


class Nil(Value):
    """Nil value."""
    type = NilType.instance()
    instance: "Nil"

    def to_str(self, context: ExecutionContext, pos: Position) -> String:
        """Convert to string."""
        return String(str(self))

    @staticmethod
    def is_truthy() -> bool:
        """Check if the value is truthy."""
        return False

    def __repr__(self):
        return "nil"


Nil.instance = Nil()


class NumberType(Type):
    """Type for number values."""
    name: str = "Number"


class Number(Value):
    """Number value."""
    type = NumberType.instance()

    def __init__(self, value: Union[int, float]):
        self.value: Union[int, float] = value

    def plus(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Plus operator."""
        other_value = other.to_num(context, pos).value
        return Number(self.value + other_value)

    def minus(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Minus operator."""
        other_value = other.to_num(context, pos).value
        return Number(self.value - other_value)

    def mul(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Multiplication operator."""
        other_value = other.to_num(context, pos).value
        return Number(self.value * other_value)

    def div(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Division operator."""
        other_value = other.to_num(context, pos).value
        if other_value == 0:
            raise ExecutionError("Cannot divide by zero", context.stack, pos)
        return Number(self.value / other_value)

    def le(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Lesser-equal operator."""
        other_value = other.to_num(context, pos).value
        return Bool.instance(self.value <= other_value)

    def lt(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Lesser-than operator."""
        other_value = other.to_num(context, pos).value
        return Bool.instance(self.value < other_value)

    def ge(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Greater-equal operator."""
        other_value = other.to_num(context, pos).value
        return Bool.instance(self.value >= other_value)

    def gt(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Greater-than operator."""
        other_value = other.to_num(context, pos).value
        return Bool.instance(self.value > other_value)

    def eq(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Equality operator."""
        if isinstance(other, Number):
            return Bool.instance(self.value == other.value)
        return Bool.false

    def neg(self, context: ExecutionContext, pos: Position) -> Value:
        """Negation unary operator."""
        return Number(-self.value)

    def to_num(self, context: ExecutionContext, pos: Position) -> Number:
        """Convert to number."""
        return self

    def to_str(self, context: ExecutionContext, pos: Position) -> String:
        """Convert to string."""
        return String(str(self))

    def __repr__(self):
        return f"{NumberType.name}({self.value})"

    def __str__(self):
        return str(self.value)


class StringType(Type):
    """Type for string values."""
    name: str = "String"


class String(Value):
    """String value."""
    type: Type = StringType.instance()

    def __init__(self, value: str):
        self.value: str = value

    def _define_attributes(self, attrs: AttributeHolder):
        super()._define_attributes(attrs)
        attrs.define(size=Const(from_py(len(self.value))))

    def plus(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Plus operator."""
        if isinstance(other, String):
            return String(self.value + other.value)
        raise ExecutionError(f"Can't concatenate {StringType.name} with {other.type.name}", context.stack, pos)

    def mul(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Multiplication operator."""
        if isinstance(other, Number):
            return String(self.value * other.value)
        raise ExecutionError(f"Can't multiply {StringType.name} and {other.type.name}", context.stack, pos)

    def le(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Lesser-equal operator."""
        if isinstance(other, String):
            return Bool.instance(self.value <= other.value)
        raise ExecutionError(f"Can't compare {StringType.name} and {other.type.name}", context.stack, pos)

    def lt(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Lesser-than operator."""
        if isinstance(other, String):
            return Bool.instance(self.value < other.value)
        raise ExecutionError(f"Can't compare {StringType.name} and {other.type.name}", context.stack, pos)

    def ge(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Greater-equal operator."""
        if isinstance(other, String):
            return Bool.instance(self.value >= other.value)
        raise ExecutionError(f"Can't compare {StringType.name} and {other.type.name}", context.stack, pos)

    def gt(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Greater-than operator."""
        if isinstance(other, String):
            return Bool.instance(self.value > other.value)
        raise ExecutionError(f"Can't compare {StringType.name} and {other.type.name}", context.stack, pos)

    def eq(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Equality operator."""
        if isinstance(other, String):
            return Bool.instance(self.value == other.value)
        return Bool.false

    def getitem(self, key: Value, context: ExecutionContext, pos: Position) -> Value:
        """Get collection element by index."""
        if isinstance(key, Number):
            return String(self.value[int(key.value)])
        raise ExecutionError(f"{String.type.name} indices must be {NumberType.name}", context.stack, pos)

    def to_str(self, context: ExecutionContext, pos: Position) -> String:
        """Convert to string."""
        return self

    def is_truthy(self) -> bool:
        """Check if the value is truthy."""
        return len(self.value) > 0

    def __str__(self):
        return self.value

    def __repr__(self):
        return f"{self.type.name}({repr(self.value)})"


class ListType(Type):
    """Type of list values."""
    name = "List"


class ListValue(Value):
    """List value."""
    type = ListType.instance()

    def __init__(self, items: List[Value]):
        self.items: List[Value] = items

    def _define_attributes(self, attrs: AttributeHolder):
        super()._define_attributes(attrs)
        attrs.define(size=Property(getter=self.size))
        attrs.define(**{
            IterableMethods.ITER: PyFunc(self.iter),
        })

    def plus(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Plus operator."""
        if isinstance(other, ListValue):
            return ListValue(self.items + other.items)
        raise ExecutionError(f"Can't add {other.type.name} to {ListType.name}", context.stack, pos)

    def mul(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Multiplication operator."""
        if isinstance(other, Number):
            return ListValue(self.items * int(other.value))
        raise ExecutionError(f"Can't multiply {ListType.name} and {other.type.name}", context.stack, pos)

    def le(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Lesser-equal operator."""
        if isinstance(other, ListValue):
            return Bool.instance(self.items <= other.items)
        raise ExecutionError(f"Can't compare {ListType.name} and {other.type.name}", context.stack, pos)

    def lt(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Lesser-than operator."""
        if isinstance(other, ListValue):
            return Bool.instance(self.items < other.items)
        raise ExecutionError(f"Can't compare {ListType.name} and {other.type.name}", context.stack, pos)

    def ge(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Greater-equal operator."""
        if isinstance(other, ListValue):
            return Bool.instance(self.items >= other.items)
        raise ExecutionError(f"Can't compare {ListType.name} and {other.type.name}", context.stack, pos)

    def gt(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Greater-than operator."""
        if isinstance(other, ListValue):
            return Bool.instance(self.items > other.items)
        raise ExecutionError(f"Can't compare {ListType.name} and {other.type.name}", context.stack, pos)

    def eq(self, other: Value, context: ExecutionContext, pos: Position) -> Value:
        """Equality operator."""
        if isinstance(other, ListValue):
            return Bool.instance(self.items == other.items)
        return Bool.false

    def getitem(self, key: Value, context: ExecutionContext, pos: Position) -> Value:
        """Get list element by index."""
        if isinstance(key, Number):
            return self.items[int(key.value)]
        raise ExecutionError(f"{ListType.name} indices must be {NumberType.name}", context.stack, pos)

    def setitem(self, key: Value, value: Value, context: ExecutionContext, pos: Position) -> Value:
        """Set list item."""
        if isinstance(key, Number):
            self.items[int(key.value)] = value
            return value
        raise ExecutionError(f"{ListType.name} indices must be {NumberType.name}", context.stack, pos)

    def to_str(self, context: ExecutionContext, pos: Position) -> String:
        """Convert to string."""
        items = ", ".join(item.to_str(context, pos).value for item in self.items)
        return String(f"[{items}]")

    def is_truthy(self) -> bool:
        """Check if the value is truthy."""
        return len(self.items) > 0

    def size(self) -> Number:
        return Number(len(self.items))

    def iter(self) -> ListIter:
        """Get iterator."""
        return ListIter(self.items)

    def __str__(self):
        items = ", ".join(str(item) for item in self.items)
        return f"[{items}]"

    def __repr__(self):
        items = ", ".join(repr(item) for item in self.items)
        return f"[{items}]"


class ListIterType(Type):
    """Type for ListIter"""
    name: str = "ListIter"


class ListIter(Value):
    """List iterator."""
    type: Type = ListIterType.instance()

    def __init__(self, items: Sequence[Value]):
        self.items: Sequence[Value] = items
        self.index: int = 0

    def _define_attributes(self, attrs: AttributeHolder):
        super()._define_attributes(attrs)
        attrs.define(**{
            IterableMethods.HAS_NEXT: PyFunc(self.has_next),
            IterableMethods.NEXT: PyFunc(self.next),
        })

    def has_next(self) -> bool:
        """Check if collection has next item."""
        return self.index < len(self.items)

    def next(self) -> Value:
        """Get next value and move iterator."""
        if self.index >= len(self.items):
            raise RuntimeError("Iterator reached the end of the list.")
        value = self.items[self.index]
        self.index += 1
        return value


def from_py(value: Any) -> Value:
    """Convert py object to Value."""
    if isinstance(value, Value):
        return value
    if isinstance(value, str):
        return String(value)
    if value is None:
        return Nil.instance
    if isinstance(value, bool):
        return Bool.instance(value)
    if isinstance(value, (int, float)):
        return Number(value)
    if isinstance(value, (list, tuple)):
        return ListValue([from_py(item) for item in value])
    raise ValueError(f"Cannot convert {type(value)} to Value")


def to_py(value: Any) -> Any:
    """Convert from lungo value to py object."""
    if not isinstance(value, Value):
        return value
    if isinstance(value, Nil):
        return None
    if isinstance(value, (String, Bool, Number)):
        return value.value
    if isinstance(value, ListValue):
        return value.items
    raise ValueError(f"Cannot convert {value.type.name} to python object")


class FunctionType(Type):
    """Type of Function values."""
    name = "Function"


class _Returned(Exception):
    """Raised by return statement."""

    def __init__(self, value: Value):
        self.value: Value = value


class ExecutableCode(abc.ABC):
    """Abstract executable code."""
    pos: Position

    @abc.abstractmethod
    def execute(self, context: ExecutionContext) -> Value:
        """Execute code."""


class BasicFunction(Value, abc.ABC):
    """Base class for user-defined function and built-in functions."""
    type = FunctionType.instance()

    def __init__(self, name: Optional[str], arg_names: Sequence[str]):
        self.name: Optional[str] = name
        self.arg_names: Sequence[str] = arg_names

    def _define_attributes(self, attrs: AttributeHolder):
        super()._define_attributes(attrs)
        attrs.define(
            name=Const(from_py(self.name)),
            args=Const(from_py(self.arg_names))
        )

    @abc.abstractmethod
    def call(self, arguments: Sequence[Value], context: ExecutionContext, pos: Position) -> Value:
        """Invoke-function operator."""

    def to_str(self, context: ExecutionContext, pos: Position) -> String:
        """Convert to string."""
        return String(repr(self))

    def __repr__(self):
        name = self.name or "func"
        args = ', '.join(self.arg_names)
        return f"{name}({args})"


class Function(BasicFunction):
    """Function value."""

    def __init__(self, name: Optional[str], arg_names: Sequence[str], body: ExecutableCode, lexical_context: Scope):
        super().__init__(name, arg_names)
        self.body: ExecutableCode = body
        self.lexical_context: Scope = lexical_context

    def call(self, arguments: Sequence[Value], context: ExecutionContext, pos: Position) -> Value:
        """Invoke-function operator."""
        if len(arguments) != len(self.arg_names):
            raise ExecutionError(f"Wrong number of arguments to call {repr(self)}", context.stack, pos)
        arguments = dict(zip(self.arg_names, arguments))
        scope = self.lexical_context.nested(arguments)
        try:
            ret_value = self.body.execute(context.push(pos, scope))
            context.stack.pop()
            return ret_value
        except _Returned as returned:
            context.stack.pop()
            return returned.value


class PyFunc(BasicFunction):
    """Wrapper around a pure python (i.e. lungo-agnostic) function."""

    def __init__(self, func: Callable, name: Optional[str] = None):
        args = tuple(inspect.signature(func).parameters.keys())
        name = name or func.__name__
        self.func: Callable = func
        super().__init__(name=name, arg_names=args)

    def call(self, arguments: Sequence[Value], context: ExecutionContext, pos: Position) -> Value:
        if len(arguments) != len(self.arg_names):
            raise ExecutionError(f"Wrong number of arguments to call {repr(self)}", context.stack, pos)
        try:
            arguments = tuple(to_py(arg) for arg in arguments)
            ret_value = self.func(*arguments)
            return from_py(ret_value)
        except Exception as e:
            raise ExecutionError(str(e), context.stack, pos)


class Operator(ExecutableCode):
    """Binary operator expression."""

    def __init__(self, operator: Callable, args: Sequence[ExecutableCode], pos: Position):
        self.operator: Callable = operator
        self.args: Sequence[ExecutableCode] = tuple(args)
        self.pos = pos

    def execute(self, context: ExecutionContext) -> Value:
        values: List[Value] = []
        for arg in self.args:
            value = arg.execute(context)
            values.append(value)
        return self.operator(context, self.pos, *values)


class Literal(ExecutableCode):
    """Constant value."""

    def __init__(self, value: Value, pos: Position):
        self.value: Value = value
        self.pos = pos

    def execute(self, context: ExecutionContext) -> Value:
        return self.value


class NameRef(ExecutableCode):
    """Name reference."""

    def __init__(self, name: str, pos: Position):
        self.name: str = name
        self.pos = pos

    def execute(self, context: ExecutionContext) -> Value:
        try:
            return context.scope.deref(self.name)
        except UndefinedName:
            raise ExecutionError(f"Undefined name '{self.name}'", context.stack, self.pos)


class Assign(ExecutableCode):
    """Assign variable."""

    def __init__(self, name: str, value: ExecutableCode, pos: Position):
        self.name: str = name
        self.value: ExecutableCode = value
        self.pos = pos

    def execute(self, context: ExecutionContext) -> Value:
        value = self.value.execute(context)
        try:
            context.scope.assign(self.name, value)
            return value
        except UndefinedName:
            raise ExecutionError(f"Undefined name '{self.name}'", context.stack, self.pos)


class Define(ExecutableCode):
    """Define variable."""

    def __init__(self, name: str, value: ExecutableCode, pos: Position):
        self.name: str = name
        self.value: ExecutableCode = value
        self.pos = pos

    def execute(self, context: ExecutionContext) -> Value:
        value = self.value.execute(context)
        try:
            context.scope.define(self.name, value)
            return value
        except NameRedefined:
            raise ExecutionError(f"Can't redefine variable '{self.name}'", context.stack, self.pos)


class Block(ExecutableCode):
    """Sequence of statements."""

    def __init__(self, statements: Sequence[ExecutableCode], pos: Position):
        self.statements: Sequence[ExecutableCode] = statements
        self.pos = pos

    def execute(self, context: ExecutionContext) -> Value:
        value = Nil.instance
        for statement in self.statements:
            value = statement.execute(context)
        return value


class FuncExpr(ExecutableCode):
    """Function expression."""

    def __init__(self, name: Optional[str], arg_names: Sequence[str], body: ExecutableCode, pos: Position):
        self.name: Optional[str] = name
        self.arg_names: Sequence[str] = arg_names
        self.body: ExecutableCode = body
        self.pos: Position = pos

    def execute(self, context: ExecutionContext) -> Value:
        func = Function(self.name, self.arg_names, self.body, context.scope)
        if self.name is not None:
            try:
                context.scope.define(self.name, func)
            except NameRedefined:
                raise ExecutionError(f"Can't redefine variable '{self.name}'", context.stack, self.pos)
        return func


class FuncCall(ExecutableCode):
    """Function invocation."""

    def __init__(self, func: ExecutableCode, args: Sequence[ExecutableCode], pos: Position):
        self.func: ExecutableCode = func
        self.args: Sequence[ExecutableCode] = args
        self.pos = pos

    def execute(self, context: ExecutionContext) -> Value:
        func = self.func.execute(context)
        args = []
        for arg in self.args:
            arg_value = arg.execute(context)
            args.append(arg_value)
        return func.call(args, context, self.pos)


class ListExpr(ExecutableCode):
    """List literal expression."""

    def __init__(self, items: Sequence[ExecutableCode], pos: Position):
        self.items: Sequence[ExecutableCode] = items
        self.pos = pos

    def execute(self, context: ExecutionContext) -> Value:
        item_values: List[Value] = []
        for item in self.items:
            value = item.execute(context)
            item_values.append(value)
        return ListValue(item_values)


class GetItem(ExecutableCode):
    """Get item expression."""

    def __init__(self, coll: ExecutableCode, key: ExecutableCode, pos: Position):
        self.coll: ExecutableCode = coll
        self.key: ExecutableCode = key
        self.pos = pos

    def execute(self, context: ExecutionContext) -> Value:
        coll = self.coll.execute(context)
        key = self.key.execute(context)
        return coll.getitem(key, context, self.pos)


class SetItem(ExecutableCode):
    """Set collection item."""

    def __init__(self, coll: ExecutableCode, key: ExecutableCode, value: ExecutableCode, pos: Position):
        self.coll: ExecutableCode = coll
        self.key: ExecutableCode = key
        self.value: ExecutableCode = value
        self.pos = pos

    def execute(self, context: ExecutionContext) -> Value:
        coll = self.coll.execute(context)
        key = self.key.execute(context)
        value = self.value.execute(context)
        return coll.setitem(key, value, context, self.pos)


class GetAttr(ExecutableCode):
    """Get attribute by name."""

    def __init__(self, value: ExecutableCode, attr: str, pos: Position):
        self.value: ExecutableCode = value
        self.attr: str = attr
        self.pos = pos

    def execute(self, context: ExecutionContext) -> Value:
        value = self.value.execute(context)
        return value.getattr(self.attr, context, self.pos)


class SetAttr(ExecutableCode):
    """Set attribute."""

    def __init__(self, target: ExecutableCode, attr: str, value: ExecutableCode, pos: Position):
        self.target: ExecutableCode = target
        self.attr: str = attr
        self.value: ExecutableCode = value
        self.pos = pos

    def execute(self, context: ExecutionContext) -> Value:
        target = self.target.execute(context)
        value = self.value.execute(context)
        target.setattr(self.attr, value, context, self.pos)
        return value


class Condition(ExecutableCode):
    """Conditional expression."""

    @dataclass
    class Branch:
        cond: ExecutableCode
        body: ExecutableCode

    def __init__(self, branches: Sequence[Branch], else_block: Optional[ExecutableCode], pos: Position):
        self.branches: Sequence[Condition.Branch] = branches
        self.else_block: Optional[ExecutableCode] = else_block
        self.pos = pos

    def execute(self, context: ExecutionContext) -> Value:
        for branch in self.branches:
            cond = branch.cond.execute(context)
            if cond.is_truthy():
                return branch.body.execute(context.nested())
        if self.else_block is not None:
            return self.else_block.execute(context.nested())


class While(ExecutableCode):
    """While loop."""

    def __init__(self, cond: ExecutableCode, body: ExecutableCode, pos: Position):
        self.cond: ExecutableCode = cond
        self.body: ExecutableCode = body
        self.pos = pos

    def execute(self, context: ExecutionContext) -> Value:
        value = Nil.instance
        while self.cond.execute(context).is_truthy():
            value = self.body.execute(context)
        return value


class IterableMethods:
    """Special method names of iterable."""
    ITER = "iter"
    HAS_NEXT = "has_next"
    NEXT = "next"


class For(ExecutableCode):
    """For loop."""

    def __init__(self, name: str, iterable: ExecutableCode, body: ExecutableCode, pos: Position):
        self.name: str = name
        self.iterable: ExecutableCode = iterable
        self.body: ExecutableCode = body
        self.pos = pos

    def execute(self, context: ExecutionContext) -> Value:
        iterable = self.iterable.execute(context)
        if not iterable.hasattr(IterableMethods.ITER, context, self.iterable.pos):
            raise ExecutionError("For loop argument must be iterable.", context.stack, self.iterable.pos)
        iterator = iterable.getattr(IterableMethods.ITER, context, self.iterable.pos).call((), context, self.pos)
        has_next = iterator.getattr(IterableMethods.HAS_NEXT, context, self.pos)
        get_next = iterator.getattr(IterableMethods.NEXT, context, self.pos)

        ret_value = Nil.instance
        context = context.nested()
        context.scope.define(self.name, Nil.instance)
        while has_next.call((), context, self.pos).is_truthy():
            iter_value = get_next.call((), context, self.pos)
            context.scope.assign(self.name, iter_value)
            ret_value = self.body.execute(context)
        return ret_value


class Return(ExecutableCode):
    """Return statement."""

    def __init__(self, value: ExecutableCode, pos: Position):
        self.value: ExecutableCode = value
        self.pos = pos

    def execute(self, context: ExecutionContext):
        value = self.value.execute(context)
        raise _Returned(value)


BuiltinFunc = Callable[[Sequence[Value], ExecutionContext, Position], Value]


def builtin(args: Sequence[str], name: Optional[str] = None) -> Callable[[BuiltinFunc], BasicFunction]:
    """Convenience decorator to create built-in function."""

    def decorator(func: BuiltinFunc) -> BasicFunction:
        func_name = name or func.__name__

        class BuiltinFunction(BasicFunction):
            def __init__(self):
                super().__init__(name=func_name, arg_names=args)

            def call(self, arguments: Sequence[Value], context: ExecutionContext, pos: Position) -> Value:
                return func(arguments, context, pos)

        return BuiltinFunction()

    return decorator



from typing import Callable, Optional, Dict

from prompt_toolkit.document import Document
from prompt_toolkit.formatted_text import StyleAndTextTuples
from prompt_toolkit.lexers import Lexer as BasePromptLexer

from lungo.lexer import Lexer, TokenType

TokeStyles = Dict[TokenType, str]


class PromptLexer(BasePromptLexer):
    DEFAULT_STYLE: TokeStyles = {
        # Special
        TokenType.SPACE: "class:pygments.whitespace",
        TokenType.NEW_LINE: "class:pygments.whitespace",
        TokenType.END: "class:pygments.whitespace",
        TokenType.UNKNOWN: "class:pygments.text",

        # Keywords:
        TokenType.FOR: "class:pygments.keyword",
        TokenType.IN: "class:pygments.keyword",
        TokenType.WHILE: "class:pygments.keyword",
        TokenType.IF: "class:pygments.keyword",
        TokenType.ELIF: "class:pygments.keyword",
        TokenType.ELSE: "class:pygments.keyword",
        TokenType.FUNC: "class:pygments.keyword",
        TokenType.RETURN: "class:pygments.keyword",
        TokenType.LET: "class:pygments.keyword",

        # Punctuation:
        TokenType.OPEN_BRACKET: "class:pygments.punctuation",
        TokenType.CLOSE_BRACKET: "class:pygments.punctuation",
        TokenType.OPEN_SB: "class:pygments.punctuation",
        TokenType.CLOSE_SB: "class:pygments.punctuation",
        TokenType.OPEN_CB: "class:pygments.punctuation",
        TokenType.CLOSE_CB: "class:pygments.punctuation",
        TokenType.COMMA: "class:pygments.punctuation",
        TokenType.SEMICOLON: "class:pygments.punctuation",

        # Operators:
        TokenType.PLUS: "class:pygments.operator",
        TokenType.MINUS: "class:pygments.operator",
        TokenType.MUL: "class:pygments.operator",
        TokenType.DIV: "class:pygments.operator",
        TokenType.AND: "class:pygments.operator",
        TokenType.OR: "class:pygments.operator",
        TokenType.NOT: "class:pygments.operator",
        TokenType.DOT: "class:pygments.operator",
        TokenType.ASSIGN: "class:pygments.operator",
        TokenType.NE: "class:pygments.operator",
        TokenType.GT: "class:pygments.operator",
        TokenType.GE: "class:pygments.operator",
        TokenType.LE: "class:pygments.operator",
        TokenType.LT: "class:pygments.operator",
        TokenType.EQ: "class:pygments.operator",

        # Atoms:
        TokenType.NAME: "class:pygments.name",
        TokenType.NUMBER: "class:pygments.number",
        TokenType.STRING: "class:pygments.string",
        TokenType.BOOL: "class:pygments.literal",
    }

    def __init__(self, lexer: Lexer, style: Optional[TokeStyles] = None):
        self.lexer: Lexer = lexer
        self.token_styles: TokeStyles = style or self.DEFAULT_STYLE

    def lex_document(self, document: Document) -> Callable[[int], StyleAndTextTuples]:
        tokens = self.lexer.tokens(text=document.text)

        lines: Dict[int, StyleAndTextTuples] = {}
        for token in tokens:
            line = lines.setdefault(token.pos.line, [])
            style = self.token_styles[token.type]
            line.append((style, token.text))

        def get_tokens(line_number: int) -> StyleAndTextTuples:
            """Get tokens by line."""

            return lines.get(line_number, [])

        return get_tokens


Prompt = Callable[[str], str]



from dataclasses import dataclass
from typing import Sequence, Optional

from lungo.lexer import Position, Token


class Node:
    """Basic node class."""
    pos: Position


class BinaryOperator(Node):
    def __init__(self, operator: Token, left: Node, right: Node, pos: Position):
        self.operator: Token = operator
        self.left: Node = left
        self.right: Node = right
        self.pos = pos


class UnaryOperator(Node):
    def __init__(self, operator: Token, arg: Node, pos: Position):
        self.operator: Token = operator
        self.arg: Node = arg
        self.pos = pos


class NumberLiteral(Node):
    def __init__(self, value: Token):
        self.value: Token = value
        self.pos = value.pos


class BoolLiteral(Node):
    def __init__(self, value: Token):
        self.value: Token = value
        self.pos = value.pos


class StringLiteral(Node):
    def __init__(self, value: Token):
        self.value: Token = value
        self.pos = value.pos


class List(Node):
    def __init__(self, items: Sequence[Node], pos: Position):
        self.items: Sequence[Node] = items
        self.pos = pos


class NameRef(Node):
    def __init__(self, name: Token):
        self.name: Token = name
        self.pos = name.pos


class FuncCall(Node):
    def __init__(self, func: Node, args: Sequence[Node], pos: Position):
        self.func: Node = func
        self.args: Sequence[Node] = tuple(args)
        self.pos = pos


class GetItem(Node):
    def __init__(self, list_expr: Node, index: Node, pos: Position):
        self.coll: Node = list_expr
        self.key: Node = index
        self.pos = pos


class Assign(Node):
    def __init__(self, name: Token, value: Node, pos: Position):
        self.name: Token = name
        self.value: Node = value
        self.pos = pos


class Block(Node):
    def __init__(self, statements: Sequence[Node], pos: Position):
        self.statements: Sequence[Node] = tuple(statements)
        self.pos = pos


class FuncExpr(Node):
    def __init__(self, name: Optional[Token], arg_names: Sequence[Token], body: Node, pos: Position):
        self.name: Optional[Token] = name
        self.arg_names: Sequence[Token] = arg_names
        self.body = body
        self.pos = pos


class DefineVar(Node):
    def __init__(self, name: Token, value: Node, pos: Position):
        self.name: Token = name
        self.value: Node = value
        self.pos = pos


class Return(Node):
    def __init__(self, value: Node, pos: Position):
        self.value: Node = value
        self.pos = pos


class Cond(Node):
    @dataclass
    class Block:
        cond: Node
        body: Node

    def __init__(self, cond_blocks: Sequence[Block], else_block: Optional[Node], pos: Position):
        self.cond_blocks = tuple(cond_blocks)
        self.else_block = else_block
        self.pos = pos


class While(Node):
    def __init__(self, cond: Node, body: Node, pos: Position):
        self.cond: Node = cond
        self.body: Node = body
        self.pos = pos


class For(Node):
    def __init__(self, name: Token, iterable: Node, body: Node, pos: Position):
        self.name: Token = name
        self.iterable: Node = iterable
        self.body: Node = body
        self.pos = pos


class GetAttr(Node):
    def __init__(self, value: Node, attr: Token, pos: Position):
        self.value: Node = value
        self.attr: Token = attr
        self.pos = pos


class SetAttr(Node):
    def __init__(self, target: Node, attr: Token, value: Node, pos: Position):
        self.target: Node = target
        self.value: Node = value
        self.attr: Token = attr
        self.pos = pos


class SetItem(Node):
    def __init__(self, coll: Node, key: Node, value: Node, pos: Position):
        self.coll: Node = coll
        self.key: Node = key
        self.value: Node = value
        self.pos = pos



from typing import Collection, Union, Sequence, Iterator

from lungo.lexer import Token, TokenType

TokenSelector = Union[TokenType, Collection[TokenType]]


class UnexpectedToken(Exception):
    def __init__(self, actual: Token, expected: TokenSelector) -> None:
        super().__init__(f"Unexpected token {actual}, expected {expected}")
        self.actual: Token = actual
        self.expected: TokenSelector = expected


def _match(token: Token, sel: TokenSelector) -> bool:
    """Match token with the selector."""
    if isinstance(sel, TokenType):
        return token.type == sel
    elif isinstance(sel, Collection):
        return token.type in sel
    else:
        raise ValueError(f"Unsupported token selector: {sel}")


def _unpack(items: Collection[Token]) -> Union[Collection[Token], Token]:
    """Replace the collection with single item by its only item."""
    if isinstance(items, Collection) and len(items) == 1:
        return next(iter(items))
    return items


class TokenStream:
    def __init__(self, tokens: Sequence[Token], position: int = 0):
        self.tokens: Sequence[Token] = tuple(tokens)
        self.position = position
        if len(self.tokens) == 0 or self.tokens[-1].type != TokenType.END:
            raise ValueError("The last token must be the token stream 'END'")

    def match(self, *selectors: TokenSelector) -> bool:
        """Check if tokens in the stream match the given selectors."""
        matched = 0
        for token, sel in zip(self, selectors):
            if not _match(token, sel):
                return False
            matched += 1
        return matched == len(selectors)

    def take(self, *selectors: TokenSelector) -> Union[Token, Sequence[Token]]:
        """Consume tokens from the stream while ensuring they match given selectors."""
        # Otherwise, take tokens if selectors are matched
        tokens = []
        for token, sel in zip(self, selectors):
            if not _match(token, sel):
                raise UnexpectedToken(token, sel)
            tokens.append(token)
        self.position = min(self.position + len(tokens), len(self.tokens) - 1)
        return _unpack(tokens)

    def __iter__(self) -> Iterator[Token]:
        """Iterate over tokens starting from the current position."""
        for i in range(self.position, len(self.tokens)):
            yield self.tokens[i]

    @property
    def current(self) -> Token:
        """Get current token."""
        return self.tokens[self.position]



import lungo.ast as ast
import lungo.runtime as rt
from lungo.lexer import TokenType, Position


class InterpreterError(Exception):
    """Indicates interpreter error."""


class StandardOperators:
    """Supported operators as static functions."""

    @staticmethod
    def plus(context: rt.ExecutionContext, pos: Position, a: rt.Value, b: rt.Value) -> rt.Value:
        return a.plus(b, context, pos)

    @staticmethod
    def minus(context: rt.ExecutionContext, pos: Position, a: rt.Value, b: rt.Value) -> rt.Value:
        return a.minus(b, context, pos)

    @staticmethod
    def mul(context: rt.ExecutionContext, pos: Position, a: rt.Value, b: rt.Value) -> rt.Value:
        return a.mul(b, context, pos)

    @staticmethod
    def div(context: rt.ExecutionContext, pos: Position, a: rt.Value, b: rt.Value) -> rt.Value:
        return a.div(b, context, pos)

    @staticmethod
    def lt(context: rt.ExecutionContext, pos: Position, a: rt.Value, b: rt.Value) -> rt.Value:
        return a.lt(b, context, pos)

    @staticmethod
    def le(context: rt.ExecutionContext, pos: Position, a: rt.Value, b: rt.Value) -> rt.Value:
        return a.le(b, context, pos)

    @staticmethod
    def gt(context: rt.ExecutionContext, pos: Position, a: rt.Value, b: rt.Value) -> rt.Value:
        return a.gt(b, context, pos)

    @staticmethod
    def ge(context: rt.ExecutionContext, pos: Position, a: rt.Value, b: rt.Value) -> rt.Value:
        return a.ge(b, context, pos)

    @staticmethod
    def eq(context: rt.ExecutionContext, pos: Position, a: rt.Value, b: rt.Value) -> rt.Value:
        return a.eq(b, context, pos)

    @staticmethod
    def and_(context: rt.ExecutionContext, pos: Position, a: rt.Value, b: rt.Value) -> rt.Value:
        return a.and_(b, context, pos)

    @staticmethod
    def or_(context: rt.ExecutionContext, pos: Position, a: rt.Value, b: rt.Value) -> rt.Value:
        return a.or_(b, context, pos)

    @staticmethod
    def neg(context: rt.ExecutionContext, pos: Position, a: rt.Value) -> rt.Value:
        return a.neg(context, pos)

    @staticmethod
    def not_(context: rt.ExecutionContext, pos: Position, a: rt.Value) -> rt.Value:
        return a.not_(context, pos)


class Translator:
    BINARY_OPERATORS = {
        TokenType.PLUS: StandardOperators.plus,
        TokenType.MINUS: StandardOperators.minus,
        TokenType.MUL: StandardOperators.mul,
        TokenType.DIV: StandardOperators.div,
        TokenType.LT: StandardOperators.lt,
        TokenType.LE: StandardOperators.le,
        TokenType.GT: StandardOperators.gt,
        TokenType.GE: StandardOperators.ge,
        TokenType.EQ: StandardOperators.eq,
        TokenType.AND: StandardOperators.and_,
        TokenType.OR: StandardOperators.or_,
    }

    UNARY_OPERATORS = {
        TokenType.MINUS: StandardOperators.neg,
        TokenType.NOT: StandardOperators.not_,
    }

    def translate(self, node: ast.Node) -> rt.ExecutableCode:
        if isinstance(node, ast.BoolLiteral):
            return rt.Literal(rt.Bool.instance(node.value.text == "true"), node.pos)
        elif isinstance(node, ast.NumberLiteral):
            return rt.Literal(rt.Number(int(node.value.text)), node.pos)
        elif isinstance(node, ast.StringLiteral):
            text = self._unescape_string(node.value.text)
            return rt.Literal(rt.String(text), node.pos)
        elif isinstance(node, ast.UnaryOperator):
            arg = self.translate(node.arg)
            return rt.Operator(self.UNARY_OPERATORS[node.operator.type], args=(arg,), pos=node.pos)
        elif isinstance(node, ast.BinaryOperator):
            left = self.translate(node.left)
            right = self.translate(node.right)
            return rt.Operator(self.BINARY_OPERATORS[node.operator.type], args=(left, right), pos=node.pos)
        elif isinstance(node, ast.Cond):
            branches = []
            for cond_block in node.cond_blocks:
                cond = self.translate(cond_block.cond)
                body = self.translate(cond_block.body)
                branches.append(rt.Condition.Branch(cond, body))
            else_block = None
            if node.else_block is not None:
                else_block = self.translate(node.else_block)
            return rt.Condition(branches, else_block, node.pos)
        elif isinstance(node, ast.While):
            cond = self.translate(node.cond)
            body = self.translate(node.body)
            return rt.While(cond, body, node.pos)
        elif isinstance(node, ast.For):
            iterable = self.translate(node.iterable)
            body = self.translate(node.body)
            return rt.For(node.name.text, iterable, body, node.pos)
        elif isinstance(node, ast.NameRef):
            return rt.NameRef(node.name.text, node.pos)
        elif isinstance(node, ast.FuncExpr):
            name = None
            if node.name is not None:
                name = node.name.text
            arg_names = [arg.text for arg in node.arg_names]
            body = self.translate(node.body)
            return rt.FuncExpr(name, arg_names, body, node.pos)
        elif isinstance(node, ast.FuncCall):
            func = self.translate(node.func)
            args = [self.translate(arg) for arg in node.args]
            return rt.FuncCall(func, args, node.pos)
        elif isinstance(node, ast.List):
            items = []
            for item in node.items:
                items.append(self.translate(item))
            return rt.ListExpr(items, node.pos)
        elif isinstance(node, ast.GetItem):
            coll = self.translate(node.coll)
            key = self.translate(node.key)
            return rt.GetItem(coll, key, node.pos)
        elif isinstance(node, ast.SetItem):
            coll = self.translate(node.coll)
            key = self.translate(node.key)
            value = self.translate(node.value)
            return rt.SetItem(coll, key, value, node.pos)
        elif isinstance(node, ast.Assign):
            value = self.translate(node.value)
            return rt.Assign(node.name.text, value, node.pos)
        elif isinstance(node, ast.DefineVar):
            value = self.translate(node.value)
            return rt.Define(node.name.text, value, node.pos)
        elif isinstance(node, ast.Return):
            value = self.translate(node.value)
            return rt.Return(value, node.pos)
        elif isinstance(node, ast.Block):
            statements = [self.translate(statement) for statement in node.statements]
            return rt.Block(statements, node.pos)
        elif isinstance(node, ast.GetAttr):
            value = self.translate(node.value)
            return rt.GetAttr(value, node.attr.text, node.pos)
        elif isinstance(node, ast.SetAttr):
            target = self.translate(node.target)
            attr = node.attr.text
            value = self.translate(node.value)
            return rt.SetAttr(target, attr, value, node.pos)
        else:
            raise InterpreterError(f"Unsupported syntax construct: {type(node)}")

    @staticmethod
    def _unescape_string(value: str) -> str:
        """Unescape string literal."""
        return value[1:-1].encode('utf-8').decode("unicode_escape")



from __future__ import annotations

import sys
from typing import Optional, TextIO, Sequence

from prompt_toolkit import PromptSession
from prompt_toolkit.input import create_input
from prompt_toolkit.output import create_output
from prompt_toolkit.styles import style_from_pygments_cls, Style
from pygments.styles.monokai import MonokaiStyle

import lungo.runtime as rt
from lungo.lexer import Lexer, Position
from lungo.parser import Parser, SyntacticError
from lungo.prompt import PromptLexer, Prompt
from lungo.sources_cache import SourceCache
from lungo.translator import Translator


def print_one(value: rt.Value):
    print(value)


class Interpreter:
    def __init__(self, lexer: Optional[Lexer] = None, parser: Optional[Parser] = None,
                 translator: Optional[Translator] = None):
        self._lexer: Lexer = lexer or Lexer()
        self._parser: Parser = parser or Parser()
        self._translator: Translator = translator or Translator()

    @staticmethod
    def make_global_scope(**custom_vars: rt.Value) -> rt.Scope:
        """Create global scope with standard library definitions."""
        return rt.Scope(symbols={
            rt.Type.name: rt.Type.instance(),
            rt.ListType.name: rt.ListType.instance(),
            rt.BoolType.name: rt.BoolType.instance(),
            rt.NumberType.name: rt.NumberType.instance(),
            rt.StringType.name: rt.StringType.instance(),
            rt.FunctionType.name: rt.FunctionType.instance(),
            rt.NilType.name: rt.NilType,
            "nil": rt.Nil.instance,
            "print": rt.PyFunc(print_one, name="print"),
            **custom_vars
        })

    @staticmethod
    def make_context(scope: rt.Scope) -> rt.ExecutionContext:
        """Create execution context."""
        return rt.ExecutionContext(scope, rt.CallStack())

    def translate_sources(self, sources: str, filename: str) -> rt.ExecutableCode:
        """Convert sources to executable code."""
        tokens = self._lexer.tokens(sources, filename)
        syntax_tree = self._parser.parse(tokens)
        return self._translator.translate(syntax_tree)

    @staticmethod
    def print_stacktrace(error: rt.ExecutionError, output: TextIO = sys.stderr, sources: Optional[SourceCache] = None):
        """Print stack-trace."""
        print("Stack trace:", file=output)
        for frame in error.stack:
            Interpreter.print_stack_entry(frame.pos, sources=sources, output=output)
        Interpreter.print_stack_entry(error.pos, sources=sources, output=output)
        print(f"{type(error).__name__}: {str(error)}")

    @staticmethod
    def print_stack_entry(pos: Position, sources: Optional[SourceCache] = None, output: TextIO = sys.stderr):
        print(f"  In file '{pos.file}', line {pos.line}", file=output)
        if sources is not None:
            line = sources.get_line(pos)
            print(f"    {line.lstrip()}", file=output)

    @staticmethod
    def print_syntactic_error(error: SyntacticError, output: TextIO = sys.stderr,
                              sources: Optional[SourceCache] = None):
        """Print syntactic error."""
        token = error.reason.actual
        pos = token.pos
        print(f"In file '{pos.file}', line {pos.line}", file=output)
        if sources is not None:
            print("  " + sources.get_line(pos), file=output)
            print("  " + " " * pos.in_line + "^" * len(token.text), file=output)
        print(f"Syntactic Error: {str(error)}", file=output)

    def make_prompt(
            self,
            prompt_lexer: Optional[PromptLexer] = None,
            style: Optional[Style] = None,
            output: TextIO = sys.stdout,
            input: TextIO = sys.stdin,
    ) -> Prompt:
        """Make prompter function."""
        prompt_session = PromptSession(
            enable_history_search=True,
            lexer=prompt_lexer or PromptLexer(self._lexer),
            style=style or style_from_pygments_cls(MonokaiStyle),
            input=create_input(input),
            output=create_output(output),
        )
        return prompt_session.prompt

    def repl(self, output: TextIO = sys.stdout, err_output: TextIO = sys.stderr, input: TextIO = sys.stdin):
        global_scope = self.make_global_scope(prompt=rt.String("lungo> "))

        sources = SourceCache()
        prompt = self.make_prompt(input=input, output=output)
        try:
            command_count = 0
            command = prompt(str(global_scope.deref("prompt")))
        except (EOFError, KeyboardInterrupt):
            return

        while True:
            try:
                command_key = f"<command:{command_count}>"
                sources.add(text=command, filename=command_key, can_evict=False)
                code = self.translate_sources(command, command_key)
                context = self.make_context(global_scope)
                value = code.execute(context)
                print(repr(value), file=output)
            except SyntacticError as error:
                self.print_syntactic_error(error, output=err_output, sources=sources)
            except rt.ExecutionError as error:
                self.print_stacktrace(error, output=err_output, sources=sources)

            try:
                command = prompt(str(global_scope.deref("prompt")))
                command_count += 1
            except (EOFError, KeyboardInterrupt):
                break

    def exec_file(self, path: str):
        """Execute file."""
        sources = SourceCache()
        try:
            text = sources.get(path)
            code = self.translate_sources(text, path)
            global_scope = self.make_global_scope()
            context = self.make_context(global_scope)
            code.execute(context)
        except SyntacticError as error:
            self.print_syntactic_error(error, sources=sources)
            sys.exit(-1)
        except rt.ExecutionError as error:
            self.print_stacktrace(error, sources=sources)
            sys.exit(-1)


def main(args: Optional[Sequence[str]] = None):
    args = args or sys.argv
    interpreter = Interpreter()
    if len(args) == 1:
        interpreter.repl()
    elif len(args) == 2:
        interpreter.exec_file(args[1])
    else:
        print("Usage:", file=sys.stderr)
        print("  lungo [FILE]", file=sys.stderr)


if __name__ == '__main__':
    main(["", "/home/stepan/PycharmProjects/parsers/main.lg"])



import enum
import re
from dataclasses import dataclass
from typing import Dict


class TokenType(enum.Enum):
    """Token types."""
    # White spaces
    SPACE = "SPACE"
    NEW_LINE = "NEW_LINE"

    # Structural
    OPEN_CB = "OPEN_CB"  # {
    CLOSE_CB = "CLOSE_CB"  # }
    OPEN_BRACKET = "OPEN_BRACKET"  # (
    CLOSE_BRACKET = "CLOSE_BRACKET"  # )
    OPEN_SB = "OPEN_SB"  # [
    CLOSE_SB = "CLOSE_SB"  # ]
    COMMA = "COMMA"
    SEMICOLON = "SEMICOLON"
    IF = "IF"
    WHILE = "WHILE"
    FOR = "FOR"
    IN = "IN"
    ELIF = "ELIF"
    ELSE = "ELSE"
    FUNC = "FUNC"
    DOT = "DOT"

    # Statements
    LET = "LET"
    RETURN = "RETURN"

    # Atomic expressions
    NUMBER = "NUMBER"
    NAME = "NAME"
    BOOL = "BOOL"
    STRING = "STRING"

    # Complex expressions
    PLUS = "PLUS"
    MINUS = "MINUS"
    MUL = "MUL"
    DIV = "DIV"
    ASSIGN = "ASSIGN"
    NOT = "NOT"
    LT = "LT"
    LE = "LE"
    GT = "GT"
    GE = "GE"
    EQ = "EQ"
    NE = "NE"
    AND = "AND"
    OR = "OR"

    # Technical
    UNKNOWN = "UNKNOWN"
    END = "END"


@dataclass
class Position:
    """Position in source code."""
    file: str  # File path
    abs: int  # Absolute position in characters
    line: int  # Line number
    in_line: int  # Position in line

    def __repr__(self):
        return f"{self.file}:{self.line}:{self.in_line}"


class Token:
    """Represents a token in a text."""

    def __init__(self, token_type: TokenType, text: str, pos: Position):
        self.type: TokenType = token_type
        self.text: str = text
        self.pos: Position = pos

    def __repr__(self):
        return f"<{self.type.name} {repr(self.text)} {self.pos}>"


class Lexer:
    """Lexical analyzer."""
    default_patterns = {
        TokenType.SPACE: r"[ \t]+",
        TokenType.NEW_LINE: r"\n",
        TokenType.SEMICOLON: r";",
        TokenType.NAME: r"(?!(?:true|false|if|elif|else|func|let|return|while|for|in)"
                        r"([^a-zA-Z_\d]|$))[a-zA-Z_][a-zA-Z_\d]*",
        TokenType.NUMBER: r"\d+",
        TokenType.OPEN_BRACKET: r"\(",
        TokenType.CLOSE_BRACKET: r"\)",
        TokenType.OPEN_CB: r"\{",
        TokenType.CLOSE_CB: r"\}",
        TokenType.OPEN_SB: r"\[",
        TokenType.CLOSE_SB: r"\]",
        TokenType.COMMA: r",",
        TokenType.PLUS: r"\+",
        TokenType.MINUS: r"\-",
        TokenType.MUL: r"\*",
        TokenType.DIV: r"/",
        TokenType.ASSIGN: r"=(?=[^=])",
        TokenType.IF: r"if",
        TokenType.ELIF: r"elif",
        TokenType.ELSE: r"else",
        TokenType.WHILE: r"while",
        TokenType.FOR: r"for",
        TokenType.IN: r"in",
        TokenType.FUNC: r"func",
        TokenType.LET: r"let",
        TokenType.RETURN: r"return",
        TokenType.BOOL: r"true|false",
        TokenType.NOT: r"!(?=[^=])",
        TokenType.LT: r"<(?=[^=])",
        TokenType.LE: r"<=",
        TokenType.GT: r">(?=[^=])",
        TokenType.GE: r">=",
        TokenType.EQ: r"==",
        TokenType.NE: r"!=",
        TokenType.AND: r"&&",
        TokenType.OR: r"\|\|",
        TokenType.DOT: r"\.",
        TokenType.STRING: r"(:?\"(:?(:?\\.)|[^\\\"])*\")|(:?'(:?(:?\\.)|[^\\\'])*')"
    }

    def __init__(self, patterns=None):
        self.patterns: Dict[TokenType, str] = patterns or Lexer.default_patterns
        regex_entries = []
        for token_type, pattern in self.patterns.items():
            regex_entries.append(f"(?P<{token_type.name}>{pattern})")
        final_regex = "|".join(regex_entries)
        self.regex: re.Pattern = re.compile(final_regex)

    def iter_tokens(self, text: str, file: str = "<code>"):
        """Split text in tokens."""
        abs_pos, line_num, line_start = 0, 0, 0
        # Iterate over known patterns
        for match in self.regex.finditer(text):
            # Get position of the found pattern match
            start, end = match.span()

            # If we skipped some text between
            # previous match and current match,
            # mark the skipped text as UNKNOWN
            if start != abs_pos:
                skipped_pos = Position(file, abs_pos, line_num, abs_pos - line_start)
                yield Token(TokenType.UNKNOWN, text[abs_pos:start], skipped_pos)

            # Produce a new token
            token_type = TokenType[match.lastgroup]
            yield Token(token_type, text[start:end], Position(file, start, line_num, start - line_start))

            # Update position
            abs_pos = end
            if token_type == TokenType.NEW_LINE:
                line_num += 1
                line_start = end

        # Mark the remaining portion of the text as UNKNOWN
        # because it doesn't contain any known patterns
        if abs_pos != len(text):
            yield Token(TokenType.UNKNOWN, text[abs_pos:], Position(file, abs_pos, line_num, abs_pos - line_start))

        # Produce the END token when we processed entire text
        yield Token(TokenType.END, '', Position(file, len(text), line_num, len(text) - line_start))

    def tokens(self, text: str, file: str = "<code>"):
        return list(self.iter_tokens(text, file))


def main():
    lexer = Lexer()
    text = "hello\n   \n\nworld$$"
    while text.strip() not in ["\\q", "exit"]:
        for token in lexer.iter_tokens(text, "<stdin>"):
            print(token)
        text = input()


if __name__ == '__main__':
    main()



from typing import Sequence, Optional

import lungo.ast as ast
from lungo.lexer import Token, TokenType, Lexer, Position
from lungo.token_stream import TokenStream, TokenSelector, UnexpectedToken


class SyntacticError(Exception):
    """Syntactic error."""

    def __init__(self, message: str, reason: UnexpectedToken) -> None:
        super().__init__(message)
        self.reason: UnexpectedToken = reason


class Parser:
    """
    Grammar:
        program = statement | statement ';' program
        statement = assign | define | ret | expr
        assign = name '=' expr
        define = 'let' name '=' expr
        ret = 'return' expr
        block = '{' program '}'
        expr = binary_expr
        binary_expr[prio=n] = binary_arg[prio=n] | binary_arg[prio=n] '<binary_operator[prio=n]>' binary_expr[prio=n]
        binary_arg[prio=n] = binary_expr[prio=n+1] | prefix_expr
        prefix_expr = '<unary_operator>' prefix_expr | postfix_expr
        postfix_expr = postfix_arg postfix_operation
        postfix_arg = ( expr ) | list | cond | while | for | func | number | bool | name
        list = '[' expr_list ']'
        postfix_operation = func_call | get_item | get_attr
        cond = 'if' '(' expr ')' block ( 'elif' '(' expr ')' block )* ('else' block )?
        while = 'while' '(' expr ')' block
        for = 'for' '(' name 'in' expr ')' block
        func = 'func' name ? '(' name_list ')' block
        func_call = '(' expr_list ')'
        get_item = '[' expr ']'
        get_attr = '.' name
        expr_list = expr | expr ',' expr_list
    """

    # Binary operators from the lowest priority to the highest
    BINARY_OPERATORS = (
        (TokenType.ASSIGN,),
        (TokenType.OR,),
        (TokenType.AND,),
        (TokenType.LT, TokenType.LE, TokenType.GT, TokenType.GE, TokenType.EQ),
        (TokenType.PLUS, TokenType.MINUS),
        (TokenType.MUL, TokenType.DIV),
    )

    # Supported unary operators
    UNARY_OPERATORS = (TokenType.MINUS, TokenType.NOT)

    def program(self, tokens: TokenStream, end_program: TokenSelector) -> ast.Node:
        """
        program = statement ';' program | statement
        """
        try:
            statements = []
            start = tokens.current.pos
            if not tokens.match(end_program):
                statement = self.statement(tokens)
                statements.append(statement)

            while not tokens.match(end_program) and not tokens.match(TokenType.SEMICOLON, end_program):
                tokens.take(TokenType.SEMICOLON)
                statement = self.statement(tokens)
                statements.append(statement)
            if tokens.match(TokenType.SEMICOLON):
                tokens.take(TokenType.SEMICOLON)
            return ast.Block(statements, pos=start)
        except UnexpectedToken as e:
            raise SyntacticError("Cannot parse code block", reason=e)

    def block(self, tokens: TokenStream) -> ast.Node:
        """
        block = '{' program '}'
        """
        try:
            tokens.take(TokenType.OPEN_CB)
            body = self.program(tokens, end_program=TokenType.CLOSE_CB)
            tokens.take(TokenType.CLOSE_CB)
            return body
        except UnexpectedToken as e:
            raise SyntacticError("Cannot parse code block", reason=e)

    def statement(self, tokens: TokenStream) -> ast.Node:
        """
        statement = define | ret | expr
        """
        if tokens.match(TokenType.LET):
            return self.define_var(tokens)
        elif tokens.match(TokenType.RETURN):
            return self.ret(tokens)
        else:
            return self.expr(tokens)

    def define_var(self, tokens: TokenStream) -> ast.Node:
        """
        define = 'let' name '=' expr
        """
        try:
            start = tokens.current.pos
            tokens.take(TokenType.LET)
            name = tokens.take(TokenType.NAME)
            tokens.take(TokenType.ASSIGN)
            value = self.expr(tokens)
            return ast.DefineVar(name, value, pos=start)
        except UnexpectedToken as e:
            raise SyntacticError("Cannot parse variable definition", reason=e)

    def ret(self, tokens: TokenStream) -> ast.Node:
        """
        ret = 'return' expr
        """
        try:
            start = tokens.current.pos
            tokens.take(TokenType.RETURN)
            value = self.expr(tokens)
            return ast.Return(value, pos=start)
        except UnexpectedToken as e:
            raise SyntacticError("Cannot parse return statement", reason=e)

    def expr(self, tokens: TokenStream) -> ast.Node:
        """
        expr = binary_expr
        """
        return self.binary_expr(tokens)

    def binary_expr(self, tokens: TokenStream, priority: int = 0) -> ast.Node:
        """
        binary_expr[prio=n] = binary_arg[prio=n] | binary_arg[prio=n] '<binary_operator[prio=n]>' binary_expr[prio=n]
        """
        start = tokens.current.pos
        expr = self.binary_arg(tokens, priority)
        while tokens.match(self.BINARY_OPERATORS[priority]):
            operator = tokens.take(self.BINARY_OPERATORS[priority])
            right_arg = self.binary_arg(tokens, priority)
            expr = self._binary_operator(operator, expr, right_arg, pos=start)
        return expr

    @staticmethod
    def _binary_operator(operator: Token, left: ast.Node, right: ast.Node, pos: Position) -> ast.Node:
        """"""
        if operator.type == TokenType.ASSIGN:
            if isinstance(left, ast.NameRef):
                return ast.Assign(name=left.name, value=right, pos=pos)
            elif isinstance(left, ast.GetItem):
                return ast.SetItem(coll=left.coll, key=left.key, value=right, pos=pos)
            elif isinstance(left, ast.GetAttr):
                return ast.SetAttr(target=left.value, attr=left.attr, value=right, pos=pos)
            else:
                raise SyntacticError(f"Cannot assign to {type(left)}", UnexpectedToken(operator, expected=()))
        return ast.BinaryOperator(operator, left, right, pos)

    def binary_arg(self, tokens: TokenStream, priority: int = 0) -> ast.Node:
        """
        binary_arg[prio=n] = binary_expr[prio=n+1] | prefix_expr
        """
        if priority + 1 < len(self.BINARY_OPERATORS):
            return self.binary_expr(tokens, priority + 1)
        return self.prefix_expr(tokens)

    def prefix_expr(self, tokens: TokenStream) -> ast.Node:
        """
        prefix_expr = '<unary_operator>' prefix_expr | postfix_expr
        """
        try:
            if tokens.match(self.UNARY_OPERATORS):
                operator = tokens.take(self.UNARY_OPERATORS)
                arg = self.prefix_expr(tokens)
                return ast.UnaryOperator(operator, arg, pos=operator.pos)
        except UnexpectedToken as e:
            raise SyntacticError("Cannot parse prefix expr", reason=e)
        return self.postfix_expr(tokens)

    def postfix_expr(self, tokens: TokenStream) -> ast.Node:
        """
        postfix_expr = postfix_arg postfix_operation
        """
        arg = self.postfix_arg(tokens)
        return self.postfix_operation(tokens, arg)

    def postfix_operation(self, tokens: TokenStream, arg: ast.Node) -> ast.Node:
        """
        postfix_operation = func_call | get_item | get_attr
        """
        if tokens.match(TokenType.OPEN_BRACKET):
            func_call = self.func_call(tokens, arg)
            return self.postfix_operation(tokens, func_call)
        elif tokens.match(TokenType.OPEN_SB):
            get_item = self.get_item(tokens, arg)
            return self.postfix_operation(tokens, get_item)
        elif tokens.match(TokenType.DOT):
            get_attr = self.get_attr(tokens, arg)
            return self.postfix_operation(tokens, get_attr)
        else:
            return arg

    def postfix_arg(self, tokens: TokenStream) -> ast.Node:
        """
        postfix_arg = '(' expr ')' | list | cond | while | for | func | number | bool | name
        """
        try:
            if tokens.match(TokenType.OPEN_BRACKET):
                tokens.take(TokenType.OPEN_BRACKET)
                expr = self.expr(tokens)
                tokens.take(TokenType.CLOSE_BRACKET)
                return expr
            elif tokens.match(TokenType.NUMBER):
                value = tokens.take(TokenType.NUMBER)
                return ast.NumberLiteral(value)
            elif tokens.match(TokenType.BOOL):
                value = tokens.take(TokenType.BOOL)
                return ast.BoolLiteral(value)
            elif tokens.match(TokenType.STRING):
                value = tokens.take(TokenType.STRING)
                return ast.StringLiteral(value)
            elif tokens.match(TokenType.NAME):
                name = tokens.take(TokenType.NAME)
                return ast.NameRef(name)
            elif tokens.match(TokenType.IF):
                return self.cond(tokens)
            elif tokens.match(TokenType.WHILE):
                return self.while_loop(tokens)
            elif tokens.match(TokenType.FOR):
                return self.for_loop(tokens)
            elif tokens.match(TokenType.FUNC):
                return self.func(tokens)
            elif tokens.match(TokenType.OPEN_SB):
                return self.list(tokens)
            else:
                expected = (
                    TokenType.OPEN_BRACKET, TokenType.NUMBER,
                    TokenType.BOOL, TokenType.NAME, TokenType.IF
                )
                raise UnexpectedToken(tokens.current, expected)
        except UnexpectedToken as e:
            raise SyntacticError("Cannot parse expression", reason=e)

    def list(self, tokens: TokenStream) -> ast.Node:
        """
        list = '[' expr_list ']'
        """
        try:
            start = tokens.current.pos
            tokens.take(TokenType.OPEN_SB)
            items = self.expr_list(tokens, list_end=TokenType.CLOSE_SB)
            tokens.take(TokenType.CLOSE_SB)
            return ast.List(items, pos=start)
        except UnexpectedToken as e:
            raise SyntacticError("Cannot parse list literal", reason=e)

    def get_item(self, tokens: TokenStream, list_expr: ast.Node) -> ast.Node:
        """
        get_item = '[' expr ']'
        """
        try:
            tokens.take(TokenType.OPEN_SB)
            index_expr = self.expr(tokens)
            tokens.take(TokenType.CLOSE_SB)
            return ast.GetItem(list_expr, index_expr, pos=list_expr.pos)
        except UnexpectedToken as e:
            raise SyntacticError("Cannot parse get-item operator", reason=e)

    def func_call(self, tokens: TokenStream, func: ast.Node) -> ast.Node:
        """
        func_call = '(' expr_list ')' | '(' ')'
        """
        try:
            tokens.take(TokenType.OPEN_BRACKET)
            args = self.expr_list(tokens, list_end=TokenType.CLOSE_BRACKET)
            tokens.take(TokenType.CLOSE_BRACKET)
            return ast.FuncCall(func, args, pos=func.pos)
        except UnexpectedToken as e:
            raise SyntacticError("Cannot parse function call operator", reason=e)

    def get_attr(self, tokens: TokenStream, value: ast.Node) -> ast.Node:
        """
        get_attr = '.' name
        """
        try:
            tokens.take(TokenType.DOT)
            name = tokens.take(TokenType.NAME)
            return ast.GetAttr(value, name, value.pos)
        except UnexpectedToken as e:
            raise SyntacticError("Cannot parse get-attr operator", reason=e)

    def expr_list(self, tokens: TokenStream, list_end: TokenSelector = TokenType.CLOSE_BRACKET) -> Sequence[ast.Node]:
        """
        expr_list = expr | expr ',' expr_list
        """
        args = []
        if not tokens.match(list_end):
            arg = self.expr(tokens)
            args.append(arg)
        while not tokens.match(list_end):
            tokens.take(TokenType.COMMA)
            arg = self.expr(tokens)
            args.append(arg)
        return args

    def func(self, tokens: TokenStream) -> ast.Node:
        """
        func = 'func' name? '(' name_list ')' block
        """
        try:
            start = tokens.current.pos
            tokens.take(TokenType.FUNC)
            name: Optional[Token] = None
            if tokens.match(TokenType.NAME):
                name = tokens.take(TokenType.NAME)
            tokens.take(TokenType.OPEN_BRACKET)
            arg_names = self.name_list(tokens, list_end=TokenType.CLOSE_BRACKET)
            tokens.take(TokenType.CLOSE_BRACKET)
            body = self.block(tokens)
            return ast.FuncExpr(name, arg_names, body, pos=start)
        except UnexpectedToken as e:
            raise SyntacticError("Cannot parse function expression", reason=e)

    def while_loop(self, tokens: TokenStream) -> ast.Node:
        """
        while = 'while' '(' expr ')' block
        """
        try:
            start = tokens.current.pos
            tokens.take(TokenType.WHILE, TokenType.OPEN_BRACKET)
            cond = self.expr(tokens)
            tokens.take(TokenType.CLOSE_BRACKET)
            body = self.block(tokens)
            return ast.While(cond, body, pos=start)
        except UnexpectedToken as e:
            raise SyntacticError("Cannot parse while loop.", reason=e)

    def for_loop(self, tokens: TokenStream) -> ast.Node:
        """
        for = 'for' '(' name 'in' expr ')' block
        """
        try:
            start = tokens.current.pos
            tokens.take(TokenType.FOR, TokenType.OPEN_BRACKET)
            name = tokens.take(TokenType.NAME)
            tokens.take(TokenType.IN)
            iterable = self.expr(tokens)
            tokens.take(TokenType.CLOSE_BRACKET)
            body = self.block(tokens)
            return ast.For(name, iterable, body, pos=start)
        except UnexpectedToken as e:
            raise SyntacticError("Cannot parse for loop.", reason=e)

    @staticmethod
    def name_list(tokens: TokenStream, list_end: TokenSelector = TokenType.CLOSE_BRACKET) -> Sequence[Token]:
        """
        name_list = name | name ',' name_list
        """
        names = []
        if not tokens.match(list_end):
            name = tokens.take(TokenType.NAME)
            names.append(name)
        while not tokens.match(list_end):
            tokens.take(TokenType.COMMA)
            name = tokens.take(TokenType.NAME)
            names.append(name)
        return names

    def cond(self, tokens: TokenStream) -> ast.Node:
        """
        cond = 'if' '(' expr ')' block ( 'elif' '(' expr ')' block )* ('else' block )?
        """
        try:
            start = tokens.current.pos
            tokens.take(TokenType.IF, TokenType.OPEN_BRACKET)
            expr = self.expr(tokens)
            tokens.take(TokenType.CLOSE_BRACKET)
            body = self.block(tokens)

            cond_blocks = [ast.Cond.Block(expr, body)]
            else_block = None

            while tokens.match(TokenType.ELIF):
                tokens.take(TokenType.ELIF)
                expr = self.expr(tokens)
                tokens.take(TokenType.CLOSE_BRACKET)
                body = self.block(tokens)
                cond_blocks.append(ast.Cond.Block(expr, body))

            if tokens.match(TokenType.ELSE):
                tokens.take(TokenType.ELSE)
                else_block = self.block(tokens)

            return ast.Cond(cond_blocks, else_block, pos=start)
        except UnexpectedToken as e:
            raise SyntacticError("Cannot parse conditional expression", reason=e)

    def parse(self, tokens: Sequence[Token]) -> ast.Node:
        ignore = (TokenType.SPACE, TokenType.NEW_LINE)
        tokens = tuple(token for token in tokens if token.type not in ignore)
        return self.program(TokenStream(tokens), end_program=TokenType.END)


def main():
    lexer = Lexer()
    parser = Parser()

    text = input()
    while text.strip() != "exit":
        tokens = lexer.tokens(text)
        syntax_tree = parser.parse(tokens)
        print("OK")
        text = input()


if __name__ == '__main__':
    main()



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Dict

from lungo.lexer import Position


class _History:
    @dataclass
    class Entry:
        key: str
        value: str
        prev: Optional[_History.Entry] = None
        next: Optional[_History.Entry] = None

        def remove(self):
            if self.prev is not None:
                self.prev.next = self.next
            if self.next is not None:
                self.next.prev = self.prev
            self.next = None
            self.prev = None

        def insert_after(self, entry: _History.Entry):
            self.next = None
            self.prev = entry
            if entry is not None:
                self.next = entry.next
                entry.next = self
                if self.next is not None:
                    self.next.prev = self

    def __init__(self):
        self._start: Optional[_History.Entry] = None
        self._end: Optional[_History.Entry] = None

    def append(self, entry: _History.Entry) -> _History.Entry:
        """Append new entry."""
        entry.insert_after(self._end)
        if self._start is None:
            self._start = entry
            self._end = entry
        return entry

    def remove(self, entry: Optional[_History.Entry]):
        """Move entry to the end of history."""
        if entry is None:
            return
        if self._start is entry:
            self._start = entry.next
        if self._end is entry:
            self._end = entry.prev
        entry.remove()

    def make_latest(self, entry: _History.Entry):
        self.remove(entry)
        self.append(entry)

    def remove_first(self) -> Optional[_History.Entry]:
        """Remove the least recent history entry."""
        first = self._start
        self.remove(first)
        return first


class SourceCache:
    def __init__(self, max_entries: int = 1000):
        self._max_entries: int = max_entries
        self._permanent: Dict[str, str] = {}
        self._cached: Dict[str, _History.Entry] = {}
        self._history: _History = _History()

    def remove(self, filename: str):
        """Remove filename from cache."""
        if filename in self._permanent:
            del self._permanent[filename]
        if filename in self._cached:
            entry = self._cached.pop(filename)
            self._history.remove(entry)

    def add(self, text: str, filename: str, can_evict: bool = True):
        """Add new sources to the cache."""
        self.remove(filename)
        if can_evict:
            entry = _History.Entry(filename, text)
            self._history.append(entry)
            self._cached[filename] = entry
            if len(self._cached) > self._max_entries:
                self.evict()
        else:
            self._permanent[filename] = text

    def get(self, filename: str) -> str:
        """Get file text."""
        if filename in self._permanent:
            return self._permanent[filename]
        elif filename in self._cached:
            entry = self._cached[filename]
            self._history.make_latest(entry)
            return entry.value
        with open(filename) as file:
            text = file.read()
            self.add(text=text, filename=filename, can_evict=True)
            return text

    def evict(self):
        """Evict the least recently used item."""
        entry = self._history.remove_first()
        if entry is not None:
            del self._cached[entry.key]

    def get_line(self, pos: Position) -> str:
        """Get line in the file."""
        text = self.get(pos.file)
        try:
            start = text.rindex("\n", 0, pos.abs) + 1
        except ValueError:
            start = 0
        try:
            end = text.index("\n", pos.abs) or len(text)
        except ValueError:
            end = len(text)
        return text[start:end]
